/************************************************************************
 * This sample demonstrates user-defined error handling functions in a
 * multithreaded application without using C++ exceptions. Although using
 * exceptions is much more convenient for a C++ programmer, there are certain
 * circumstances (especially on embedded platforms) when exceptions are
 * undesirable, e.g., for performance reasons. As said in the ASN.1/C++
 * Runtime Reference, the default error handling is unsuitable for
 * multithreaded applications. The below sample replaces the default
 * error handling with a custom error handling that does not use exceptions
 * but is multithreading-aware.
 ***************************************************************************/

#include "bcasm.h"

/***************************************************************************
 * The sample is designed to be run both on Windows and on Unix-like OSes
 * supporting pthreads. Windows threads and pthreads use rather different
 * APIs. So we define some macros to isolate the dependency on a particular
 * multithreading system. The rest of the sample uses the macros and so is
 * independent of the OS.
 ***************************************************************************/

#if defined(_WIN32)
#include <process.h>
#define THREAD		HANDLE
#define THREADFUNC	unsigned __stdcall
#define TLS_TYPE	DWORD
#define CREATE_THREAD(th, sub, parm)	{ unsigned dummy; th = (HANDLE)_beginthreadex(NULL, 0, &sub, parm, 0, &dummy); }
#define WAIT_THREAD(th)	WaitForSingleObject(th, INFINITE)
#define EXIT_THREAD	{ _endthreadex(0); return 0; }
#define ALLOC_TLS	TlsIndex = TlsAlloc()
#define GET_TLS		TlsGetValue(TlsIndex)
#define SET_TLS(value)	TlsSetValue(TlsIndex, value)
#define DELETE_TLS	TlsFree(TlsIndex)
#else
#include <pthread.h>
#include <unistd.h>
#define THREAD		pthread_t
#define THREADFUNC	void *
#define TLS_TYPE	pthread_key_t
#define CREATE_THREAD(th, sub, parm)	{ if (pthread_create(&th, NULL, &sub, parm)) th = 0; }
#define WAIT_THREAD(th)	pthread_join(th, NULL)
#define EXIT_THREAD	return NULL
#define ALLOC_TLS	pthread_key_create(&TlsIndex, NULL)
#define GET_TLS		pthread_getspecific(TlsIndex)
#define SET_TLS(value)	pthread_setspecific(TlsIndex, value)
#define DELETE_TLS	pthread_key_delete(TlsIndex)
#endif

/***************************************************************************
 * The sample, after the needed initialization, starts a certain number of
 * worker threads. Each worker thread encodes some PDU value (namely, myData
 * value defined in the source ASN.1 file), deliberately corrupts the
 * resulting encoding (to cause a decoding error), and tries to decode the
 * changed encoding. These operations are repeated a certain number of times,
 * in a loop. Each thread changes a specific byte in the encoding, so errors
 * in the same thread must have the same error code, which may be checked in
 * the sample output. For simplicity, the sample prints messages using simple
 * printf. This may cause output lines produced by different threads interfere.
 * In a more real situation, this interferention should be avoided, probably
 * by posting error messages to a special thread, which prints them in order.
 ***************************************************************************/

#define NUM_OF_THREADS	300		/* total number of worker threads */
#define ENCDECS_PER_THREADS  1000       /* the number of encoding/decodings
					   in each thread */

TLS_TYPE TlsIndex;		/* TLS slot to store the thread identifier */
int error_codes[NUM_OF_THREADS];/* array of error codes, element per a thread */

/* Error handler. This function replaces the default ASN.1/C++ error handler */
void my_error_handler(int code)
{
    /* Stores the error code in the corresponding element of the array of
       error codes. The index of the element depends on the current thread. */
    *(int *)GET_TLS = code;
}

/* Error checking function. This function returns the error code of the last
   operation and cleans the error code. Thus, it combines the functionality
   of asn1_get_last_error() and asn1_clean_error(). This is not necessary,
   you may implement two functions instead. It is just an alternative
   approach to the default error handling.
*/
int my_get_error()
{
    int *ptr = (int *)GET_TLS;
    int code;

    code = *ptr;
    *ptr = 0;
    return code;
}

/* The thread function */
THREADFUNC threadfunc (void *parm)
{
    int threadid = (int *)parm - error_codes;
    int i;
    int errcode;

    /* Store the pointer to the thread's element in the array of error codes
       in the TLS. In a real situation, the structure pointed to by the TLS
       slot may contain extra information needed for the application. */
    SET_TLS(parm);

    /* Now we are using the new error handling */
    bcasm_Control ctl;
    if ((errcode = my_get_error())) {
	printf("Thread %d - initialization error; code %d\n", threadid, errcode);
	EXIT_THREAD;
    }
    ctl.setEncodingFlags(ctl.getEncodingFlags() | NOTRAPPING);
    if ((errcode = my_get_error())) {
	printf("Thread %d - cannot set flags; code %d\n", threadid, errcode);
	EXIT_THREAD;
    }
    ctl.setDecodingFlags(ctl.getDecodingFlags() | NOTRAPPING);
    if ((errcode = my_get_error())) {
	printf("Thread %d - cannot set flags; code %d\n", threadid, errcode);
	EXIT_THREAD;
    }
    EncodedBuffer encodedData;
    if ((errcode = my_get_error())) {
     	printf("Thread %d - cannot create encoded data buffer; code %d\n",
		threadid, errcode);
	EXIT_THREAD;
    }
    BBCard data = myCard;
    if ((errcode = my_get_error())) {
	printf("Thread %d - error %d copying data\n", threadid, errcode);
	EXIT_THREAD;
    }
    BBCard_PDU pdu;
    for (i = 0; i < ENCDECS_PER_THREADS; i++) {
	/* Encoding data */
        pdu.set_data(data);
	pdu.encode(ctl, encodedData);
	if ((errcode = my_get_error())) {
	    printf("Thread %d - error %d encoding data\n", threadid, errcode);
	    EXIT_THREAD;
	}

	/* Change one byte in the encoded data (to cause an error, possibly) */
	encodedData.get_data()[threadid % encodedData.get_length()]++;

	/* Decoding. Errors are expected here */
	pdu.decode(ctl, encodedData);
	if ((errcode = my_get_error())) {
	    /* Decoding failed */
	    printf("Thread %d - error %d decoding data\n", threadid, errcode);
	} else {
	    /* Decoding succeeded - free the decoded data */
	    pdu.free_data(ctl);
	}
	/* Free the encoded data */
	encodedData.set_buffer(0, NULL);
    }
    printf("Thread %d finished\n", threadid);
    EXIT_THREAD;
}

/* Main program */
int main()
{
    int i;
    THREAD threads[NUM_OF_THREADS];	/* Here we store the IDs of worker threads */

    /* Set the new error handling */
    asn1_set_error_handling(my_error_handler, TRUE);

    /* Allocate the TLS slot to use for the new error handling */
    ALLOC_TLS;

    /* Start threads */
    for (i = 0; i < NUM_OF_THREADS; i++) {
	CREATE_THREAD(threads[i], threadfunc, &error_codes[i]);
	if (threads[i] == 0) {
	    printf("Cannot start thread No. %d - finishing the test.\n", i);
	}
    }

    /* Wait for threads to finish */
    for (i = 0; i < NUM_OF_THREADS; i++) {
	if (threads[i] != 0)
	    WAIT_THREAD(threads[i]);
    }

    /* Delete the allocated TLS slot */
    DELETE_TLS;
    return 0;
}
